#include "display.h"

#include <math.h>

#ifdef WIN32
#include <windows.h> // A bizarre dependency for GL/gl.h
#endif

#ifdef __APPLE__
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#endif

// Constructor
Display::Display()
: displayNormals(false)
{
	std::cout << "Display::Display()" << std::endl;
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_NORMALIZE);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glClearDepth(1.0f);
	std::cout << "end Display::Display()" << std::endl;
}

// Destructor
Display::~Display()
{
	// Release textures loaded into OpenGL
	std::list<unsigned int>::iterator i;
	for (i = textures.begin(); i != textures.end(); i++)
		glDeleteTextures(1, &*i);
}

// Produce perspective projection matrix
void Display::perspective(float fov, float aspect, float zNear, float zFar)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	float f = static_cast<float>(1 / tan(fov / 2 * 3.1415926535 / 180.0));
	float zDiff = zNear - zFar;
	float m[16] =
	{
		f / aspect, 0, 0, 0,
		0, f, 0, 0,
		0, 0, (zFar + zNear) / zDiff, -1,
		0, 0, 2 * zFar * zNear / zDiff, 0
	};
	glMultMatrixf(m);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

// Produce orthogonal projection matrix
void Display::ortho(float fleft, float fright, float fbottom, float ftop, float fnear, float ffar)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(fleft, fright, fbottom, ftop, fnear, ffar);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

// Toggle depth testing
void Display::depthTest(bool depthTest)
{
	if (depthTest)
		glEnable(GL_DEPTH_TEST);
	else
		glDisable(GL_DEPTH_TEST);
}

// Resize the viewport
void Display::resize(int x, int y)
{
	std::cout << "Display::resize(" << x << ", " << y << ")" << std::endl;
	windowWidth = x;
	windowHeight = y;
	if (!y) y = 1;
	float fov = float(x) / float(y);
	glViewport(0, 0, x, y);
	perspective(90.0f, fov, 0.1f, 300.0f);
}

// Display a mesh, using vertex arrays
void Display::display(Mesh &mesh)
{
	if (mesh.vertices.size())
	{
		glVertexPointer(3, GL_FLOAT, 0, reinterpret_cast<GLvoid *>(&*mesh.vertices.begin()));
		glEnableClientState(GL_VERTEX_ARRAY);
	}
	if (mesh.color.size())
	{
		glColorPointer(3, GL_FLOAT, 0, reinterpret_cast<GLvoid *>(&*mesh.color.begin()));
		glEnableClientState(GL_COLOR_ARRAY);
	}
	if (mesh.normal.size())
	{
		glNormalPointer(GL_FLOAT, 0, reinterpret_cast<GLvoid *>(&*mesh.normal.begin()));
		glEnableClientState(GL_NORMAL_ARRAY);
	}
	if (mesh.texture.size())
	{
		glTexCoordPointer(2, GL_FLOAT, 0, reinterpret_cast<GLvoid *>(&*mesh.texture.begin()));
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	if (mesh.triangle.size())
	{
		glDrawElements(GL_TRIANGLES, static_cast<GLsizei>(mesh.triangle.size()), GL_UNSIGNED_INT, reinterpret_cast<GLvoid *>(&*mesh.triangle.begin()));
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	if (displayNormals)
	{
		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		glBegin(GL_LINES);
		for (size_t i = 0; i < mesh.triangle.size(); i += 3)
		{
			vertex &v = mesh.vertices[mesh.triangle[i]];
			vertex &n = mesh.normal[mesh.triangle[i]];
			glColor3f(1.0f, 0.0f, 0.0f);
			glVertex3f(v.x, v.y, v.z);
			glColor3f(0.0f, 1.0f, 0.0f);
			glVertex3f(v.x + n.x, v.y + n.y, v.z + n.z);
		}
		glEnd();
		glPopAttrib();
	}
}

/* Display the terrain.  This probably shouldn't be here.  I think we'd be better off
	creating a true mesh from the terrain data that can be displayed with a vertex array and
	sending that to this display class. */
void Display::display(Terrain &terrain)
{
	glColor3f(1.0f, 1.0f, 1.0f);

	unsigned int width = terrain.width();
	unsigned int height = terrain.height();
	unsigned int x;
	unsigned int y;
	vertex scale = terrain.scale();
	unsigned int lastTexture = 0;

	for (x = 0; x < width - 1; x++)
	{
		for (y = 0; y < height - 1; y++)
		{
			TerrainValue tv[2][2];
			tv[0][0] = terrain.at(x, y);
			tv[0][1] = terrain.at(x, y + 1);
			tv[1][0] = terrain.at(x + 1, y);
			tv[1][1] = terrain.at(x + 1, y + 1);
			unsigned int v[6][2] = {{x, y}, {x, y + 1}, {x + 1, y + 1}, {x + 1, y + 1}, {x + 1, y}, {x, y}};
			TerrainValue h[6];
			h[0] = tv[0][0];
			h[1] = tv[0][1];
			h[2] = tv[1][1];
			h[3] = tv[1][1];
			h[4] = tv[1][0];
			h[5] = tv[0][0];

			int orientation = terrain.at(x, y).orientation;
			float texCoords[4][6][2] // [orientation][i][dimension]
			= {
				{{1,0}, {1,1}, {0,1}, {0,1}, {0,0}, {1,0}},
				{{1,1}, {0,1}, {0,0}, {0,0}, {1,0}, {1,1}},
				{{0,1}, {0,0}, {1,0}, {1,0}, {1,1}, {0,1}},
				{{0,0}, {1,0}, {1,1}, {1,1}, {0,1}, {0,0}}
				};
			
			if (terrain.at(x, y).texture != lastTexture)
			{
				select(terrain.at(x, y).texture);
				lastTexture = terrain.at(x, y).texture;
			}
			glBegin(GL_TRIANGLES);
			
			for (int i = 0; i < 6; i++)
			{
				glTexCoord2f(texCoords[orientation][i][0], texCoords[orientation][i][1]);
				glNormal3f(h[i].normal.x, h[i].normal.y, h[i].normal.z);
				glVertex3f(scale.x * float(v[i][0]) / (width - 1), scale.y * h[i].height, scale.z * float(v[i][1]) / (height - 1));
			}

			glEnd();
		}
	}

	if (displayNormals)
	{
		glDisable(GL_LIGHTING);
		glBegin(GL_LINES);
		glColor3f(0.0f, 1.0f, 1.0f);
		for (x = 0; x < width; x++)
		{
			for (y = 0; y < height; y++)
			{
				float f = 1.0f;
				glColor3f(1.0f, 0.0f, 0.0f);
				glVertex3f(scale.x * float(x) / width, scale.y * terrain.at(x, y).height, scale.z * float(y) / height);
				glColor3f(0.0f, 1.0f, 0.0f);
				glVertex3f(scale.x * float(x) / width + terrain.at(x, y).normal.x * f, scale.y * terrain.at(x, y).height + terrain.at(x, y).normal.y * f, scale.z * float(y) / height + terrain.at(x, y).normal.z * f);
			}
		}
		glEnd();
		glEnable(GL_LIGHTING);
	}
}

// Display a console.
void Display::display(Console &console)
{
	// Assert(left != right && top != bottom)
	float left = 0.0f;
	float right = 1.0f;
	float bottom = 0.5f;
	float top = 1.0f;

	// We need a nice way to get these values in here as non-magic numbers.
	int fontWidth = 256;
	int fontHeight = 256;

	std::list<std::string>::iterator i;
	int charWidth = fontWidth / 16;
	int charHeight = fontHeight / 16;
	int numRows = int((top - bottom) * (windowHeight / charHeight));
	int numCols = int((right - left) * (windowWidth / charWidth));
	int col = 0;
	int row = 0;
	float width = float(charWidth);
	float height = float(charHeight);
	std::list<std::string> &buffer = console.data();

	left *= windowWidth;
	right *= windowWidth;
	bottom *= windowHeight;
	top *= windowHeight;

	glDisable(GL_DEPTH_TEST);

	glDisable(GL_TEXTURE_2D);
	glColor4f(0.3f, 0.3f, 0.3f, 0.3f);
	glBegin(GL_QUADS);
	glVertex2f(left, bottom);
	glVertex2f(right, bottom);
	glVertex2f(right, top);
	glVertex2f(left, top);
	glEnd();
	glEnable(GL_TEXTURE_2D);

	glColor3f(1.0f, 1.0f, 1.0f);

	for (i = buffer.begin(); i != buffer.end() && row < numRows; i++)
	{
		unsigned int rowsRequired = static_cast<unsigned int>(i->size() / numCols);
		int startRow = row;
		row += rowsRequired;
		std::string::iterator p;

		p = i->begin();
		if (row >= numRows)
		{
			p += (row - numRows + 1) * numCols;
			startRow = row = numRows - 1;
		}

		for (; p != i->end(); p++)
		{
			float charSize = 1.0f / 16;
			int x = *p % 16;
			int y = 15 - *p / 16;

			int l = int(left + col * width);
			int r = int(l + width);
			int t = int(bottom + row * height);
			int b = int(t + height);

			glBegin(GL_QUADS);
			glTexCoord2f(x * charSize, (y + 1) * charSize);
			glVertex2f(static_cast<float>(l), static_cast<float>(b));
			glTexCoord2f((x + 1) * charSize, (y + 1) * charSize);
			glVertex2f(static_cast<float>(r), static_cast<float>(b));
			glTexCoord2f((x + 1) * charSize, y * charSize);
			glVertex2f(static_cast<float>(r), static_cast<float>(t));
			glTexCoord2f(x * charSize, y * charSize);
			glVertex2f(static_cast<float>(l), static_cast<float>(t));
			glEnd();

			col++;
			if (col >= numCols)
			{
				col = 0;
				row--;
			}
		}
		row = startRow + rowsRequired + 1;
		col = 0;
	}

	glEnable(GL_DEPTH_TEST);
}

// Display overlayed menus
void Display::display(MenuData &menu)
{
	// We need a nice way to get these values in here as non-magic numbers.
	int fontWidth = 512;
	int fontHeight = 512;

	int charWidth = fontWidth / 16;
	int charHeight = fontHeight / 16;
	float width = float(charWidth);
	float height = float(charHeight);

	unsigned int maxHeight = static_cast<unsigned int>(menu.data.size());
	unsigned int maxWidth = 0;
	for (unsigned int item = 0; item < menu.data.size(); item++)
		if (menu.data[item].label.size() > maxWidth)
			maxWidth = static_cast<unsigned int>(menu.data[item].label.size());

	float useWidth = float(charWidth * maxWidth);
	float useHeight = float(charHeight * maxHeight);
	float left = windowWidth / 2.0f  - useWidth / 2.0f;
	float bottom = windowHeight / 2.0f - useHeight / 2.0f;

	unsigned int numRows = static_cast<unsigned int>(menu.data.size());

	glDisable(GL_DEPTH_TEST);

	glColor3f(1.0f, 1.0f, 1.0f);

	float charSize = 1.0f / 16;

	for (size_t row = 0; row < menu.data.size(); row++)
	{
		std::string label = menu.data[row].label;
		if (menu.data[row].choices.size())
		{
			label += ": ";
			label += menu.data[row].choices[menu.data[row].selected];
		}
		int col = 0;
		for (std::string::iterator i = label.begin(); i != label.end(); i++)
		{
			int x = *i % 16;
			int y = 15 - *i / 16;

			int l = int(left + col * width);
			int r = int(l + width);
			int t = int(bottom + (numRows - row - 1) * height);
			int b = int(t + height);

			if (row == menu.selected)
				glColor3f(1.0f, 0.0f, 0.0);
			else
				glColor3f(1.0f, 1.0f, 1.0f);

			glBegin(GL_QUADS);
			glTexCoord2f(x * charSize, (y + 1) * charSize);
			glVertex2f(static_cast<float>(l), static_cast<float>(b));
			glTexCoord2f((x + 1) * charSize, (y + 1) * charSize);
			glVertex2f(static_cast<float>(r), static_cast<float>(b));
			glTexCoord2f((x + 1) * charSize, y * charSize);
			glVertex2f(static_cast<float>(r), static_cast<float>(t));
			glTexCoord2f(x * charSize, y * charSize);
			glVertex2f(static_cast<float>(l), static_cast<float>(t));
			glEnd();

			col++;
		}
	}

	glEnable(GL_DEPTH_TEST);
}

// Display a skeleton
void Display::display(SkeletonNode &skeleton, std::map<std::string, Mesh> &meshTable)
{
	glPushMatrix();
	if (skeleton.translation.x || skeleton.translation.y || skeleton.translation.z)
		glTranslatef(skeleton.translation.x, skeleton.translation.y, skeleton.translation.z);
	if (skeleton.rotation.x)
		glRotatef(skeleton.rotation.x, 1.0f, 0.0f, 0.0f);
	if (skeleton.rotation.y)
		glRotatef(skeleton.rotation.y, 0.0f, 1.0f, 0.0f);
	if (skeleton.rotation.z)
		glRotatef(skeleton.rotation.z, 0.0f, 0.0f, 1.0f);
	if (skeleton.scale.x != 1.0f || skeleton.scale.y != 1.0f || skeleton.scale.z != 1.0f)
		glScalef(skeleton.scale.x, skeleton.scale.y, skeleton.scale.z);
	display(meshTable[skeleton.mesh]);
	for (std::vector<SkeletonNode>::iterator i = skeleton.child.begin(); i != skeleton.child.end(); i++)
		display(*i, meshTable);
	glPopMatrix();
}

// Display an object
void Display::display(Object &object, std::map<std::string, Mesh> &meshTable)
{
	select(object.texture);
	glPushMatrix();
	glTranslatef(object.position.x, object.position.y, object.position.z);
	glScalef(object.scale.x, object.scale.y, object.scale.z);
	if (object.rotation.x) glRotatef(object.rotation.x, 1.0f, 0.0f, 0.0f);	
	if (object.rotation.y) glRotatef(object.rotation.y, 0.0f, 1.0f, 0.0f);	
	if (object.rotation.z) glRotatef(object.rotation.z, 0.0f, 0.0f, 1.0f);	
	display(object.skeleton, meshTable);
	glPopMatrix();
}

// Toggle drawing of vertex normals
void Display::normals(bool b)
{
	displayNormals = b;
}

unsigned int Display::load(Image &texture)
{
	unsigned int mode = 0;
	unsigned int id = 0;
	switch (texture.bpp)
	{
	case 8: mode = GL_ALPHA; break; // I think this should work, but we really need to test it.
	case 24: mode = GL_RGB; break;
	case 32: mode = GL_RGBA; break;
	default: return 0; // We don't care about any other formats.
	}
	glGenTextures(1, &id);
	glBindTexture(GL_TEXTURE_2D, id);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, mode, texture.width, texture.height, 0, mode, GL_UNSIGNED_BYTE, &*texture.data.begin());
	textures.push_back(id); // Keep track of all textures loaded so that we can unload them politely.
	return id;
}

// Bind a texture
void Display::select(unsigned int texture)
{
	if (texture)
		glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);
}

// Relase a texture
void Display::release(unsigned int texture)
{
	std::list<unsigned int>::iterator i;
	 
	 for (i = textures.begin(); i != textures.end(); i++)
		if (*i == texture)
		{
			textures.erase(i);
			break;
		}
	glDeleteTextures(1, &texture);
}

// Draw some simple axes.  This was useful at one point.  I swear.
void Display::axes()
{
	glBegin(GL_LINES);
	glColor3f(1.0f, 0.0f, 0.0f);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(1.0f, 0.0f, 0.0f);
	glColor3f(0.0f, 1.0f, 0.0f);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(0.0f, 1.0f, 0.0f);
	glColor3f(0.0f, 0.0f, 1.0f);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(0.0f, 0.0f, 1.0f);
	glEnd();
}

// Draw a rectangle.  How quaint.
void Display::rectangle(float left, float right, float bottom, float top)
{
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(left, bottom);
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(right, bottom);
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(right, top);
	glTexCoord2f(0.0f, 1.0f);
	glVertex2f(left, top);
	glEnd();
}

// Simple text output.
void Display::text(const char *message, float left, float bottom, float charWidth, float charHeight)
{
	char ch;
	glBegin(GL_QUADS);
	while ((ch = *message++))
	{
		float charSize = 1.0f / 16;
		int x = ch % 16;
		int y = 15 - ch / 16;

		glTexCoord2f(x * charSize, y * charSize);
		glVertex2f(left, bottom);
		glTexCoord2f((x + 1) * charSize, y * charSize);
		glVertex2f(left + charWidth, bottom);
		glTexCoord2f((x + 1) * charSize, (y + 1) * charSize);
		glVertex2f(left + charWidth, bottom + charHeight);
		glTexCoord2f(x * charSize, (y + 1) * charSize);
		glVertex2f(left, bottom + charHeight);
		left += charWidth;
	}
	glEnd();
}

// Toggle a light
void Display::lightEnable(int lightNumber, bool lightEnabled)
{
	if (lightEnabled)
		glEnable(GL_LIGHT0 + lightNumber);
	else
		glDisable(GL_LIGHT0 + lightNumber);
}

// Toggle all lighting
void Display::lighting(bool lightEnabled)
{
	if (lightEnabled)
		glEnable(GL_LIGHTING);
	else
		glDisable(GL_LIGHTING);
}

// Set the light's position
void Display::lightPosition(int lightNumber, const vertex &position)
{
	GLfloat p[4] = {position.x, position.y, position.z, 1.0f};
	glLightfv(GL_LIGHT0 + lightNumber, GL_POSITION, p);
}

// Set diffuse properties of light
void Display::lightDiffuse(int lightNumber, const vertex &diffuse)
{
	GLfloat d[4] = {diffuse.x, diffuse.y, diffuse.z, 1.0f};
	glLightfv(GL_LIGHT0 + lightNumber, GL_DIFFUSE, d);
}

// Toggle fog
void Display::fog(bool enabled)
{
	if (enabled)
	{
		glEnable(GL_FOG);
		glHint(GL_FOG_HINT, GL_DONT_CARE);
		glFogi(GL_FOG_MODE, GL_EXP2);
	}
	else
		glDisable(GL_FOG);
}

// Set fog color
void Display::fogColor(float red, float green, float blue, float density)
{
	float v[4] = {red, green, blue, 1.0f};
	glFogfv(GL_FOG_COLOR, v);
	glFogf(GL_FOG_DENSITY, density);
}

// Set fog cutoff
void Display::fogCutoff(float start, float end)
{
	glFogf(GL_FOG_START, start);
	glFogf(GL_FOG_END, end);
}

// Clear depth and color buffers
void Display::clear()
{
	glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
}

// Set the camera location
void Display::camera(const vertex &position, const vertex &rotation)
{
	glLoadIdentity();
	glRotatef(rotation.x, 1.0f, 0.0f, 0.0f);
	glRotatef(rotation.y, 0.0f, 1.0f, 0.0f);
	glRotatef(rotation.z, 0.0f, 0.0f, 1.0f);
	glTranslatef(-position.x, -position.y, -position.z);
}

// Toggle wireframe mode
void Display::wireframe(bool b)
{
	glPolygonMode(GL_FRONT_AND_BACK, b?GL_LINE:GL_FILL);
}

// Set color
void Display::color(float red, float green, float blue)
{
	glColor3f(red, green, blue);
}
